import datetime
from typing import List

import pandas as pd
from teachingtoolshed.gradebook.csv_readers import CSVReader


class Canvas:
    def __init__(
        self,
        filename: str,
        student_name_col: str = "Student",
        sid_col: str = "SIS Login ID",
        dummy_rows: int = 2,
        out_dir: str = "out",
    ):
        """Class that manages state and changes to a Canvas Gradebook export.

        Args:
            filename: A filename to a Canvas Gradebook export to use as the starting grades.
            student_name_col: Column name in Gradebook that contains the student name
            sid: Column name in Gradebook that contains students UW Net ID (or whatever you will
              use to identify students)
            dummy_rows: Optional; The number of rows to skip in the Gradebook before the column
              header rows. These rows will be preserved in the output but need to parsed separately
              from the students and regular column headers.
        """
        self.filename: str = filename
        self.student_name_col: str = student_name_col
        self.sid_col: str = sid_col
        self.dummy_rows: int = dummy_rows
        self.out_dir = out_dir

        # Useful to know which assignments we changed for reporting differences
        self.changes: List[str] = []

        # Read in data
        df = pd.read_csv(filename)
        self.original_canvas: pd.DataFrame = df[self.dummy_rows :]
        self.dummies: pd.DataFrame = df[: self.dummy_rows]
        self.canvas: pd.DataFrame = self.original_canvas.set_index(self.sid_col)

    def _find_column(self, col_name_prefix: str, grab_first: bool = False) -> str:
        """Given the prefix of a column name, returns a full column name that matches this prefix.

        Useful for specifying a short prefix of a name rather than the whole Canvas Column
        (which contains an arbitrary identifier). If the prefix does not uniquely identify one
        column, will raise an error (unless grab_first is True, in which case it returns the first)

        Args:
            col_name_prefix: The prefix of a column name
            grade_first: Determines behavior in the case where more than one column has
              col_name_prefix. If True, returns the first. If False, raises an ValueError.
        """
        if col_name_prefix in self.canvas:
            return col_name_prefix
        else:
            columns = self.canvas.columns
            potential_columns = columns[columns.str.startswith(col_name_prefix)]
            if len(potential_columns) == 1 or (
                grab_first and len(potential_columns) > 1
            ):
                return potential_columns[0]
            else:
                raise ValueError(
                    f"canvas_col ({col_name_prefix}) does not uniquely identify one column"
                )

    def add_grades(
        self, canvas_col_name: str, csv_reader: CSVReader, grab_first: bool = False
    ):
        """Adds the scores from csv_reader to the Gradebook column canvas_col_name.

        The CSVReader and its subclasses read in student grades from other sources and
        are indexed by some unique student identifier (e.g., UW NetID). Note the values
        in the csv_reader's identifiers must match the sid values in this class.

        Args:
            canvas_col_name: The prefix or full name of a column in the Canvas Gradebook
            csv_reader: A CSVReader containing scores from another source
            grade_first: Determines behavior in the case where more than one column has
              col_name_prefix. If True, returns the first. If False, raises an ValueError.
        """
        canvas_col_name = self._find_column(canvas_col_name, grab_first=grab_first)

        self.changes.append(canvas_col_name)

        # Join together to get scores for this column
        self.canvas[canvas_col_name] = self.canvas.join(csv_reader.scores, how="left")[
            csv_reader.score_col
        ]

        # Give everyone else a 0
        self.canvas[canvas_col_name].fillna(0, inplace=True)

    def set_grade(
        self,
        student_id: str,
        canvas_col_name: str,
        score: float,
        grab_first: bool = False,
    ):
        """
        Sets the grade for the given student for the given assignment to a particular value
        """
        canvas_col_name = self._find_column(canvas_col_name, grab_first=grab_first)
        self.canvas.loc[student_id, canvas_col_name] = score

    def get_grade(
        self, student_id: str, canvas_col_name: str, grab_first: bool = False
    ):
        canvas_col_name = self._find_column(canvas_col_name, grab_first=grab_first)
        return self.canvas.loc[student_id, canvas_col_name]

    def report_diffs(self, verbose=False):
        """Utility method to report differences for assignments added"""

        df_merged = self.original_canvas.merge(
            self.canvas,
            left_on=self.sid_col,
            right_on=self.sid_col,
            suffixes=("_old", "_new"),
        )

        # Find rows that changed
        diffs = None
        for changed_col in self.changes:
            col_diffs = df_merged[changed_col + "_old"].astype(float) != df_merged[
                changed_col + "_new"
            ].astype(float)

            if verbose and col_diffs.any():
                print(f"Changes for {changed_col}")
                print(
                    df_merged.loc[
                        col_diffs,
                        [
                            self.student_name_col + "_old",
                            self.sid_col,
                            changed_col + "_old",
                            changed_col + "_new",
                        ],
                    ]
                )

            if diffs is None:
                diffs = col_diffs
            else:
                diffs = diffs | col_diffs

        if diffs.sum() == 0:
            print("No differences found! 😱")
        else:
            print(f"Found {diffs.sum()} differences")
            cols = (
                [self.student_name_col + "_old", self.sid_col]
                + [col + "_old" for col in self.changes]
                + [col + "_new" for col in self.changes]
            )

            if verbose and diffs.sum() > 0:
                print("All Changes")
                print(df_merged.loc[diffs, cols])
            return df_merged[diffs][cols].copy()

    def export(self, filename: str = None):
        """Saves the current Gradebook to a new filename

        Args:
            filename: Optional; File name to save to. If None,
              generates a timestamped filename.
        """
        if filename is None:
            filename = self.export_filename()

        df = self.dummies.append(
            self.canvas.reset_index(), ignore_index=True, sort=False
        )
        df.to_csv(filename, index=False)

    def export_filename(self) -> str:
        """Returns a timestamped filename for exporting"""
        now = datetime.datetime.now()
        date_format = "%Y-%b-%d-at-%H-%M"
        return f"{self.out_dir}/Canvas-Export-{now.strftime(date_format)}.csv"



from typing import Any, Callable, Dict, Hashable, List, Mapping, Optional, Union

import pandas as pd

# Simple CSV Readers that read a single score column from various formatted CSVs


class CSVReader:
    def __init__(
        self,
        filename: str,
        sid_col: str,
        score_col: Union[str, List[str]],
        sid_is_email: bool = False,
        dummy_rows: int = 0,
    ):
        """Reads in a CSV containing scores for an assignment.

        Almost any grading tool outputs a CSV of scores for an assignment. This class
        is responsible for part of the process of moving the scores from an export CSV
        to your course gradebook (i.e., Canvas). Some systems store UW emails instead of
        UW NetID, so sid_is_email controls if the trailing part of the email ('@uw.edu')
        should be removed.

        Args:
            filename: Filename containing scores for an assignment
            sid_col: Name of the column in this CSV that identifies a student (e.g., a UW NetID)
            score_col: Name of the column in this CSV that has student scores for this assignment or a list of columns
            sid_is_email: Optional; If True, removes the suffix of the sid_col values related to an
              email address (i.e., '@uw.edu')
            dummy_rows: Optional; Number of rows to skip in the CSV at the beginning
        """
        self.filename: str = filename
        self.dummy_rows: int = dummy_rows
        self.score_col: Union[str, List[str]] = score_col
        self.sid_col: str = sid_col  # Might modify after reading in data

        # Read in data
        self.scores: pd.DataFrame = pd.read_csv(filename, skiprows=self.dummy_rows)

        # Change sid_col to store just UW Net IDs if they are emails
        if sid_is_email:
            self.scores[self.sid_col] = self.scores[self.sid_col].str.split(
                "@", expand=True
            )[0]

        # Make the data frame indexed by net id
        self.scores = self.scores.set_index(self.sid_col)

        # Drop all columns that aren't the score column. Keep as DataFrame
        if type(score_col) is str:
            score_columns = [score_col]
        else:  # type is list
            score_columns = score_col

        self.scores = self.scores[score_columns]


class GradescopeReader(CSVReader):
    def __init__(
        self,
        filename: str,
        sid_col: str,
        score_col: Union[str, List[str]] = "Total Score",
        dummy_rows: int = 0,
    ):
        """Helper class for common type of CSV export. See documentation for CSVReader"""
        super().__init__(filename, sid_col, score_col, dummy_rows=dummy_rows)


class EdStemReader(CSVReader):
    def __init__(
        self,
        filename: str,
        sid_col: str,
        score_col: Union[str, List[str]] = "feedback grade",
        sid_is_email: bool = True,
        dummy_rows: int = 0,
        rename_index: Dict[str, str] = {},
    ):
        """Helper class for common type of CSV export. See documentation for CSVReader.

        EdStem stores emails rather than a student id (by default). Some students use their
        personal email as their account email so we use rename_index to map these few
        students to their proper student ID

        Args:
            See CSVReader for most arguments
            rename_index: dict from personal email (without email suffix) to correct student ID
        """
        super().__init__(
            filename,
            sid_col,
            score_col,
            sid_is_email=sid_is_email,
            dummy_rows=dummy_rows,
        )
        self.scores = self.scores.rename(index=rename_index)  # type: ignore


# More complicated CSV Readers



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['teachingtoolshed', 'teachingtoolshed.api', 'teachingtoolshed.gradebook']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.5.0,<2.0.0', 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'teaching-toolshed',
    'version': '0.1.3',
    'description': 'Helpful libraries for running classes',
    'long_description': '# Teaching Toolshed\n\nA library of useful classes for writing scripts related to managing classes. Has helper classes for talking to various teaching APIs and for compiling gradebooks in Python. \n\nSee [Teaching Toolshed Examples](https://github.com/hschafer/teaching-toolshed-examples) to see example scripts of how these libraries are used in my classes.\n\nAny questions or comments can be handled here on GitHub or you can email me at [hschafer@cs.washington.edu](mailto:hschafer@cs.washington.edu).\n\n## Publish New Version\n\nSee [here](https://realpython.com/pypi-publish-python-package/#publish-your-package-to-pypi)\n',
    'author': 'Hunter Schafer',
    'author_email': 'hschafer@uw.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Module defining access to the EdStem API

Note that the EdStem API is still under development and there is no promise for
its stability. Use at your own risk.

Every request requires you authenticate. In order to do this, you will need your
authentication token from EdStem. You can access your token by looking at network requests
on EdStem and finding a request with an x-token header.
"""
import json
from typing import Any, Dict, List

import requests
import itertools

# Special type to indicate only a 0 or 1 should be passed
BinaryFlag = int


def urljoin(*parts):
    """Combines parts of a URL into a fully path.

    Removes any additional trailing or leading "/" characters.

    Example:
      urljoin('abc.com', 'path/', 'file.txt) -> 'abc.com/path/file.txt'

    Args:
        *parts: Any sequence of parts of a URL to join

    Returns:
        Concatenate the parts of the URL together, separated by "/".
    """
    return "/".join(str(part).strip("/") for part in parts)


class EdStemAPI:
    API_URL = f"https://us.edstem.org/api/"

    def __init__(self, course_id: int, token: str):
        """Initializes access to the EdStem API for a course with the given ID.


        Args:
            course_id: An integer course ID for the course to access. See EdStem URL.
            token: Your EdStem authentication token
        """
        self._course_id = course_id
        self._token = token

    # General functions for GET/POST
    def _ed_get_request(
        self, url: str, query_params: Dict[str, Any] = {}
    ) -> Dict[str, Any]:
        """Sends a GET request to EdStem.

        Args:
            url: URL endpoint to hit
            query_params: A dictionary of query parameters and values

        Returns:
            A JSON response from the endpoint

        Raises:
            HTTPError: If there was an error with the HTTP request
        """
        response = requests.get(
            url, params=query_params, headers={"Authorization": "Bearer " + self._token}
        )
        response.raise_for_status()
        return response.json()

    def _ed_post_request(
        self, url: str, query_params: Dict[str, Any] = {}, json: Dict[str, Any] = {}
    ) -> bytes:
        """Sends a POST request to EdStem.

        Args:
            url: URL endpoint to hit
            query_params: A dictionary of query parameters and values
            json: A dictionary of parameters and values to pass as the payload

        Returns:
            A binary string containing response content

        Raises:
            HTTPError: If there was an error with the HTTP request
        """
        response = requests.post(
            url,
            params=query_params,
            json=json,
            headers={"Authorization": "Bearer " + self._token},
        )
        response.raise_for_status()
        return response.content

    def _ed_put_request(
        self,
        url: str,
        query_params: Dict[str, Any] = {},
        json: Dict[str, Any] = {},
        data: Dict[str, Any] = {},
    ) -> bytes:
        """Sends a PUT request to EdStem.

        Args:
            url: URL endpoint to hit
            query_params: A dictionary of query parameters and values
            json: A dictionary of parameters and values to pass as the payload

        Returns:
            A binary string containing response content

        Raises:
            HTTPError: If there was an error with the HTTP request
        """
        response = requests.put(
            url,
            params=query_params,
            json=json,
            data=data,
            headers={"Authorization": "Bearer " + self._token},
        )
        response.raise_for_status()
        return response.content

    def _ed_delete_request(
        self,
        url: str,
        query_params: Dict[str, Any] = {},
        json: Dict[str, Any] = {},
        data: Dict[str, Any] = {},
    ) -> bytes:
        """Sends a DELETE request to EdStem.

        Args:
            url: URL endpoint to hit
            query_params: A dictionary of query parameters and values
            json: A dictionary of parameters and values to pass as the payload

        Returns:
            A binary string containing response content

        Raises:
            HTTPError: If there was an error with the HTTP request
        """
        response = requests.delete(
            url,
            params=query_params,
            json=json,
            data=data,
            headers={"Authorization": "Bearer " + self._token},
        )
        response.raise_for_status()
        return response.content

    # Enrollment info
    def get_users(self):
        admin_path = urljoin(EdStemAPI.API_URL, f"courses/{self._course_id}/admin")
        admin_info = self._ed_get_request(admin_path)
        return admin_info["users"]

    # Get lesson info
    def get_all_lessons(self) -> List[Dict[str, Any]]:
        """Gets all lessons for a course. Endpoint: /courses/{course_id}/lessons

        Returns:
            A list of JSON objects, one for each lesson.
        """
        lessons_path = urljoin(EdStemAPI.API_URL, f"courses/{self._course_id}/lessons")
        lessons = self._ed_get_request(lessons_path)
        return lessons["lessons"]

    # Get module info
    def get_all_modules(self) -> List[Dict[str, Any]]:
        """Gets all modules for a course. Endpoint: /courses/{course_id}/lessons

        Returns:
            A list of JSON objects, one for each module.
        """
        lessons_path = urljoin(EdStemAPI.API_URL, f"courses/{self._course_id}/lessons")
        lessons = self._ed_get_request(lessons_path)
        return lessons["modules"]

    def get_lesson(self, lesson_id: int) -> Dict[str, Any]:
        """Gets metadata for a single lesson. Endpoint: /lessons/{lesson_id}

        Args:
            lesson_id: Identifier for lesson

        Returns:
            A JSON object with the specified lesson's metadata
        """
        lesson_path = urljoin(EdStemAPI.API_URL, "lessons", lesson_id)
        lesson = self._ed_get_request(lesson_path)["lesson"]
        return lesson

    def get_slide(self, slide_id: int) -> Dict[str, Any]:
        """Gets metadata for a single slide. Endpoint: /lessons/slides/{slide_id}

        Args:
            slide_id: Identifier for slide

        Returns:
            A JSON object with the specified slide's metadata
        """
        slide_path = urljoin(EdStemAPI.API_URL, "lessons", "slides", slide_id)
        slide = self._ed_get_request(slide_path)["slide"]
        return slide

    def create_lesson(
        self, title: str = None, options: Dict[str, Any] = {}
    ) -> Dict[str, Any]:
        """Creates a new Ed lesson. Endpoint: /courses/{course_id}/lessons

        Args:
            title: Title for the new lesson
            options: Dictionary of options to set on the lesson

        Returns:
            A JSON object with the new lesson's metadata
        """
        lessons_path = urljoin(EdStemAPI.API_URL, f"courses/{self._course_id}/lessons")
        lesson_dict = {"lesson": ({"title": title} | options)}
        lesson = json.loads(self._ed_post_request(lessons_path, json=lesson_dict))[
            "lesson"
        ]
        return lesson

    def edit_lesson(
        self, lesson_id: int, options: Dict[str, Any] = {}
    ) -> Dict[str, Any]:
        """Modifies an existing Ed lesson. Endpoint: /lessons/{lesson_id}

        Args:
            lesson_id: Identifier for lesson
            options: Dictionary of options to set on the lesson

        Returns:
            A JSON object with the updated lesson's metadata
        """
        lesson = self.get_lesson(lesson_id)
        lesson_path = urljoin(EdStemAPI.API_URL, f"lessons/{lesson_id}")
        lesson_dict = {"lesson": lesson | options}
        lesson = json.loads(self._ed_put_request(lesson_path, json=lesson_dict))[
            "lesson"
        ]
        return lesson

    def clone_slide(
        self,
        slide_id: int,
        lesson_id: int,
        is_hidden: bool = False,
    ) -> Dict[str, Any]:
        """Clones an existing Ed slide into a new lesson. Endpoint: /lessons/slides/{slide_id}/clone

        Args:
            slide_id: Identifier for slide to clone
            lesson_id: Identifier for lesson to clone into

        Returns:
            A JSON object with the cloned slide's metadata
        """

        clone_path = urljoin(EdStemAPI.API_URL, f"lessons/slides/{slide_id}/clone")
        payload = {"lesson_id": lesson_id, "is_hidden": is_hidden}
        slide = json.loads(self._ed_post_request(clone_path, json=payload))["slide"]
        return slide

    def edit_slide(self, slide_id: int, options: Dict[str, Any] = {}) -> Dict[str, Any]:
        """Modifies an existing Ed slide. Endpoint: /lessons/slides/{slide_id}

        Args:
            slide_id: Identifier for slide
            options: Dictionary of options to set on the slide

        Returns:
            A JSON object with the updated slide's metadata
        """
        slide = self.get_slide(slide_id)
        slide_path = urljoin(EdStemAPI.API_URL, f"lessons/slides/{slide_id}")
        slide_dict = slide | options
        slide = json.loads(
            self._ed_put_request(slide_path, data={"slide": json.dumps(slide_dict)})
        )["slide"]
        return slide

    def get_questions(self, slide_id: int) -> List[Dict[str, Any]]:
        """Gets metadata for a single Quiz slide's questions. Endpoint: /lessons/slides/{slide_id}/questions

        Args:
            slide_id: Identifier for slide

        Returns:
            A JSON object with the specified lesson's metadata
        """
        questions_path = urljoin(
            EdStemAPI.API_URL, "lessons", "slides", slide_id, "questions"
        )
        questions = self._ed_get_request(questions_path)["questions"]
        return questions

    def edit_question(
        self, question_id: int, question_data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Edits the question with the given id with the updated question data.

        Endpoint: /lessons/slides/questions/{question_id}

        Args:
            question_id: Identifier for question
            question_data: Dict containing information to update in the queston

        Returns:
            The updated JSON object for the question
        """
        update_question_path = urljoin(
            EdStemAPI.API_URL, "lessons", "slides", "questions", question_id
        )

        # API call requires the data to be the value of a "question" key
        if "question" not in question_data:
            question_data = {"question": question_data}

        response = self._ed_put_request(update_question_path, json=question_data)
        return json.loads(response)["question"]

    def delete_question(self, question_id: int) -> None:
        """Deletes the given question with this id. Endpoint /lessons/slides/questions/{question_id}

        Args:
            question_id: Identifier for question

        Returns:
            None
        """
        delete_path = urljoin(
            EdStemAPI.API_URL, "lessons", "slides", "questions", question_id
        )
        self._ed_delete_request(delete_path)

    # Methods for getting information about lesson/assignment completion
    def get_lesson_completions(
        self,
        lesson_id: int,
        completions: BinaryFlag = 1,
        numbers: BinaryFlag = 1,
        scores: BinaryFlag = 1,
        students: BinaryFlag = 1,
        strategy: str = "latest",
        ignore_late: BinaryFlag = 0,
        late_no_points: BinaryFlag = 0,
        tz: str = "America/Los_Angeles",
    ) -> bytes:
        """Gets completion information for a single lesson. Endpoint: /lessons/{lesson_id}/results.csv

        Same as using the "Download Lesson Results..." (completions=0) or
        "Download Lesson Completions..." (completions=1) option in the lesson menu.
        Our documentation is in terms of the options seen in this popup. Checkboxes correspond to
        BinaryFlag values 1 for checked and 0 for unchecked.

        Args:
            lesson_id: Identifier for lesson
        Optional Args:
            completions: If 1, gets completions and if 0 gets results. Completions show timestamps
              for when each slide was finished. Results shows scores.
            numbers: Check; "Slide numbers: Include a header row with the number of each slide"
            scores: Check; "Slide scores: Include a header row with the score for each slide"
            students: Check; "Include students only: Only student results will be included in the report"
            strategy: Indicate which submissions should be downloaded
              Currently supports: 'best' and 'latest'
            ignore_late: Check; "Ignore late submissions"
            late_no_points: Check; "Late submissions are worth 0 points"
            tz: Timezone for datetimes

        Returns:
            Bytes content of the result file. Usually will be used to save to a file.
        """
        lesson_completion_path = urljoin(
            EdStemAPI.API_URL, "lessons", lesson_id, "results.csv"
        )
        result = self._ed_post_request(
            lesson_completion_path,
            {
                "numbers": numbers,
                "scores": scores,
                "students": students,
                "completions": completions,
                "strategy": strategy,
                "ignore_late": ignore_late,
                "late_no_points": late_no_points,
                "tz": tz,
            },
        )
        return result

    def get_challenge_results(
        self,
        challenge_id: int,
        students: BinaryFlag = 1,
        feedback: BinaryFlag = 0,
        type: str = "optimised",
        score_type: str = "pertestcase",
        numbers: BinaryFlag = 0,
        scores: BinaryFlag = 0,
        tz: str = "America/Los_Angeles",
    ) -> bytes:
        """Gets results for a single coding challenge. Endpoint: /challenges/{challenge_id}/results.csv

        Note that both lesson coding challenges (including the Jupyter type) and assignments
        are both considered coding challenges. The challenge_id is NOT the same as the slide_id for a
        challenge slide.

        Same as using the "Download Results..." assessments menu.
        Our documentation is in terms of the options seen in this popup. Checkboxes correspond to
        BinaryFlag values 1 for checked and 0 for unchecked.

        Args:
            challenge_id: Identifier for challenge (not the same as a slide_id)
        Optional Args:
            students: Check; "Include students only"
            feedback: Check; "Include feedback criteria columns and comment"
            type: Dropdown; "Choose which submissions will be included in the report"
                Options: 'latest', 'latest-with-feedback', 'optimised, 'all'
            numbers: Currently not specifiable in the UI
            scores: Currently not specifiable in the UI
            score_type: Radio; "Choose how testcases scores are reported"
                Options: 'pertestcase' and 'passfail'
            tz: Timezone for datetimes

        Returns:
            Bytes content of the result file. Usually will be used to save to a file.
        """
        challenge_path = urljoin(
            EdStemAPI.API_URL, "challenges", challenge_id, "results"
        )
        result = self._ed_post_request(
            challenge_path,
            {
                "students": students,
                "type": type,
                "numbers": numbers,
                "scores": scores,
                "score_type": score_type,
                "feedback": feedback,
                "tz": tz,
            },
        )
        return result

    def get_quiz_results(
        self, quiz_id: int, students: BinaryFlag = 1, no_attempt: BinaryFlag = 1
    ):
        """Gets results for a single quiz. Endpoint: /lessons/slides/{quiz_id}/questions/results

        Note that the quiz_id is not the same as the slide_id for a quiz slide

        Same as using the "Download Quiz Responses..." quiz slide menu.
        Our documentation is in terms of the options seen in this popup. Checkboxes correspond to
        BinaryFlag values 1 for checked and 0 for unchecked.

        Args:
            challenge_id: Identifier for challenge (not the same as a slide_id)
        Optional Args:
            students: Check; "Include students only: Only student responses will be included in the report"
            no_attempt: Check; "Show empty attempts: Create a row for the user in the results even if the user has not attempted the quiz"

        Returns:
            Bytes content of the result file. Usually will be used to save to a file.

        """
        quiz_path = urljoin(
            EdStemAPI.API_URL, "lessons/slides", quiz_id, "questions/results"
        )
        result = self._ed_post_request(
            quiz_path,
            {
                "students": students,
                "noAttempt": no_attempt,
            },
        )
        return result

    def post_grades(
        self,
        submission_id: int,
        grades: List[Dict[str, Any]],
        comment: str = '<document version="2.0"><paragraph/></document>',
    ):
        """
        Posts feedback to the given submission id with the given grades and optional comment.

        Args:
            submission_id: ID for submission
            grades: List of marks to apply, one per criteria (see example below).
        Optiona Args:
            comments: Contents for the overall feedback. **Note: Must be a proper XML document**

        Example:
        >>> ed.post_grades(123,
        >>>     [[{name: "Stacks/Queues", description: "Excellent", mark: "E"}],
        >>>     '<document version="2.0"><paragraph>Nice job!</paragraph></document>')
        """
        path = urljoin(
            EdStemAPI.API_URL, "challenges", "submissions", submission_id, "feedback"
        )

        data = {
            "feedback": {
                "mark": None,
                "formatted": True,
                "criteria": grades,
                "content": comment,
            }
        }
        return self._ed_put_request(path, json=data)

    def connect_user_to_workspace(self, challenge_id, user_id):
        connect_path = urljoin(EdStemAPI.API_URL, "challenges", challenge_id, "connect")
        self._ed_post_request(connect_path, json={"user_id": user_id})

    def submit_all_challenge(self, challenge_id):
        submit_path = urljoin(
            EdStemAPI.API_URL, "challenges", challenge_id, "submit_all"
        )
        self._ed_post_request(submit_path)

    def submit_all_quiz(self, slide_id):
        submit_path = urljoin(
            EdStemAPI.API_URL, "lessons", "slides", slide_id, "questions", "submit_all"
        )
        self._ed_post_request(submit_path)

    def get_all_users_for_challenge(self, challenge_id):
        users_path = urljoin(EdStemAPI.API_URL, "challenges", challenge_id, "users")

        result = self._ed_get_request(users_path)["users"]
        return result

    def get_all_users(self):
        users = self._ed_get_request(
            urljoin(EdStemAPI.API_URL, "courses", self._course_id, "analytics", "users")
        )["users"]
        return users

    def get_all_tutorials(self):
        users = self.get_all_users()
        groups = itertools.groupby(
            sorted(users, key=lambda x: x["tutorial"] if x["tutorial"] else ""),
            key=lambda x: x["tutorial"],
        )

        tutorials = []
        for k, _ in groups:
            tutorials.append(k)
        return tutorials

    def get_all_submissions(
        self,
        challenge_id: int,
        students: BinaryFlag = 1,
        type: str = "optimised",
        tz: str = "America/Los_Angeles",
    ):

        # TODO also add ability to specify before date
        submission_path = urljoin(
            EdStemAPI.API_URL, "challenges", challenge_id, "submissions"
        )
        result = self._ed_post_request(
            submission_path,
            query_params={"studuents": students, "type": type, "tz": tz},
        )
        return result

    def get_all_submissions_for_user(
        self,
        challenge_id,
        user_id,
    ):
        submission_path = urljoin(
            EdStemAPI.API_URL,
            "users",
            user_id,
            "challenges",
            challenge_id,
            "submissions",
        )

        result = self._ed_get_request(submission_path)["submissions"]
        return result

    def delete_submission(self, sub_id):
        delete_path = urljoin(EdStemAPI.API_URL, "challenges", "submissions", sub_id)
        return self._ed_delete_request(delete_path)
